import fs from 'fs/promises';
import path from 'path';
import cpy from 'cpy';
import merge from 'deepmerge';
import { globby } from 'globby';
import Handlebars from 'handlebars';
import { JSDOM } from 'jsdom';
import { format } from 'prettier';
import { registerBuiltInHelpers } from './helpers/built-in.js';
import { AfterAllPluginError, AfterTransformPluginError, AmbiguousPartialError, ConfigError, ReadError, TransformPluginError, WriteError } from './utilities/errors.js';
import { getRelativeUrl, getDestinationFilename, getPathname } from './utilities/files.js';
import { parse as parseMarkdown, render as renderMarkdown } from './utilities/markdown.js';
import { render } from './utilities/render.js';
import { parseUrl, urlToPlainObject } from './utilities/urls.js';
export * from './types.js';
export const defaultConfig = {
    source: '.',
    destination: './_site',
    baseUrl: '/',
    layouts: 'layouts',
    partials: 'partials',
    assets: '',
    sharedScripts: 'shared/scripts.js',
    sharedStyles: 'shared/styles.css',
    data: {},
    cleanBeforePublishing: false,
    plugins: [],
    helpers: [],
    pageExtension: 'html',
    filterPages: () => true
};
export default class Doxicity {
    constructor(config) {
        this.hasPublished = false;
        this.hasRegisteredPartials = false;
        this.config = { ...defaultConfig, ...config };
        this.config.source = path.resolve(this.config.source);
        this.config.destination = path.resolve(this.config.destination);
        registerBuiltInHelpers(this.config);
        this.config.helpers.forEach(helper => {
            Handlebars.registerHelper(helper.name, helper.callback);
        });
    }
    async registerPartials() {
        const partialDir = path.resolve(this.config.source, this.config.partials);
        const filenames = await globby([path.join(partialDir, '**/*.{hbs,md}')]);
        const partials = filenames.map(filename => ({
            name: path.basename(filename.toLowerCase(), path.extname(filename.toLowerCase())),
            filename
        }));
        let partial = { name: '', filename: '' };
        const duplicates = partials.filter(p => {
            const found = partials.filter(f => f.name === p.name);
            return found.length > 1 ? true : false;
        });
        if (duplicates.length) {
            const message = `Doxicity found ambiguous partials. This can happen when partial files have the same name but different ` +
                `extensions.\n\n` +
                `${duplicates.map(d => d.filename).join('\n')}\n\n` +
                `Please rename these files and try again.\n`;
            throw new AmbiguousPartialError(duplicates.map(d => d.filename), message);
        }
        try {
            for await (partial of partials) {
                const content = await fs.readFile(partial.filename, 'utf-8');
                const html = renderMarkdown(content, { preserveHandlebars: true });
                Handlebars.registerPartial(partial.name, html);
            }
        }
        catch (err) {
            const message = `Doxicity couldn't read the following partial from the file system.\n\n` +
                `${partial.filename}\n\n` +
                `Please verify file permissions and try again.\n`;
            throw new ReadError(message);
        }
    }
    async generateSharedScripts() {
        const filename = path.join(this.config.destination, this.config.sharedScripts);
        let scripts = '';
        for await (const plugin of this.config.plugins) {
            if (plugin.scripts) {
                const js = await plugin.scripts(this.config);
                scripts += `
          /* ${plugin.name} */
          (function() {
            ${js}
          })();
        `;
            }
        }
        try {
            scripts = format(scripts, { parser: 'babel' });
        }
        catch (err) {
        }
        if (scripts.trim().length > 0) {
            try {
                await fs.mkdir(path.dirname(filename), { recursive: true });
                await fs.writeFile(filename, scripts, 'utf8');
            }
            catch (err) {
                const message = `Doxicity couldn't write the following file to the file system.\n\n` +
                    `${filename}\n\n` +
                    `Please verify that the target directory has the correct permissions and try again.\n`;
                throw new WriteError(message);
            }
            return getRelativeUrl(this.config, filename);
        }
        return null;
    }
    async generateSharedStyles() {
        const filename = path.join(this.config.destination, this.config.sharedStyles);
        let styles = '';
        for await (const plugin of this.config.plugins) {
            if (plugin.styles) {
                const css = await plugin.styles(this.config);
                styles += `
          /* ${plugin.name} */
          ${css}
        `;
            }
        }
        try {
            styles = format(styles, { parser: 'css' });
        }
        catch (err) {
        }
        if (styles.trim().length > 0) {
            try {
                await fs.mkdir(path.dirname(filename), { recursive: true });
                await fs.writeFile(filename, styles, 'utf8');
            }
            catch (err) {
                const message = `Doxicity couldn't write the following file to the file system.\n\n` +
                    `${filename}\n\n` +
                    `Please verify that the target directory has the correct permissions and try again.\n`;
                throw new WriteError(message);
            }
            return getRelativeUrl(this.config, filename);
        }
        return null;
    }
    static resolve(...args) {
        return path.resolve(...args);
    }
    async publish(options) {
        const opts = {
            copyAssets: true,
            ...options
        };
        const publishStartTime = Date.now();
        const ignoredDirectories = [
            path.join(this.config.source, this.config.layouts),
            path.join(this.config.source, this.config.partials)
        ];
        const publishedPages = [];
        const sharedScripts = await this.generateSharedScripts();
        const sharedStyles = await this.generateSharedStyles();
        const files = (await globby([
            path.resolve(this.config.source, '**/*.md'),
            ...ignoredDirectories.map(dir => `!${path.resolve(this.config.source, dir)}/**/*.md`)
        ])).filter(file => this.config.filterPages(file));
        if (!this.hasRegisteredPartials) {
            await this.registerPartials();
        }
        if (this.config.cleanBeforePublishing) {
            await this.clean();
        }
        for await (const file of files) {
            const sourceFile = file;
            const destinationFile = getDestinationFilename(sourceFile, this.config);
            const pathname = getPathname(sourceFile, this.config);
            const { frontMatter, content } = await parseMarkdown(sourceFile);
            const data = merge(this.config.data, frontMatter);
            const url = urlToPlainObject(parseUrl(this.config.baseUrl));
            Object.keys(data).forEach(key => {
                if (key.startsWith('$')) {
                    const message = `A field named "${key}" was found in your config data. Fields cannot start with $ because those names ` +
                        `are reserved for magic properties.\n`;
                    throw new ConfigError(message);
                }
            });
            data.$content = content;
            data.$url = url;
            const page = {
                sourceFile,
                destinationFile,
                pathname,
                data
            };
            let html = await render(page, this.config);
            publishedPages.push(page);
            const doc = new JSDOM(html, {
                url: `https://internal/`
            }).window.document;
            if (sharedScripts) {
                const script = doc.createElement('script');
                script.src = sharedScripts;
                doc.body.append(script);
            }
            if (sharedStyles) {
                const firstStylesheet = doc.head.querySelector('link[rel="stylesheet"]');
                const link = doc.createElement('link');
                link.rel = 'stylesheet';
                link.href = sharedStyles;
                if (firstStylesheet) {
                    firstStylesheet.insertAdjacentElement('beforebegin', link);
                }
                else {
                    doc.head.prepend(link);
                }
            }
            for (const plugin of this.config.plugins) {
                if (plugin.transform) {
                    try {
                        await plugin.transform(doc, page, this.config);
                    }
                    catch (err) {
                        const message = `A plugin called "${plugin.name} threw an error during the "transform" phase of the following page.\n\n` +
                            `${page.sourceFile}\n\n` +
                            `Here's some more info about the error.\n\n` +
                            `${err.message}\n`;
                        throw new TransformPluginError(plugin.name, page, message);
                    }
                }
            }
            html = `<!DOCTYPE html>\n${doc.documentElement.outerHTML}`;
            for (const plugin of this.config.plugins) {
                if (plugin.afterTransform) {
                    try {
                        html = await plugin.afterTransform(html, page, this.config);
                    }
                    catch (err) {
                        const message = `A plugin called "${plugin.name} threw an error during the "afterTransform" phase of the following page.\n\n` +
                            `${page.sourceFile}\n\n` +
                            `Here's some more info about the error.\n\n` +
                            `${err.message}\n`;
                        throw new AfterTransformPluginError(plugin.name, page, message);
                    }
                    if (!html) {
                        const message = `A plugin called "${plugin.name} failed to return any HTML during the "afterTransform" phase of the following page.\n\n` +
                            `${page.sourceFile}\n\n` +
                            `Please update the plugin and try again.\n`;
                        throw new AfterTransformPluginError(plugin.name, page, message);
                    }
                }
            }
            try {
                await fs.mkdir(path.dirname(page.destinationFile), { recursive: true });
                await fs.writeFile(page.destinationFile, html, 'utf-8');
            }
            catch (err) {
                const message = `Doxicity couldn't write the following file to the file system.\n\n` +
                    `${page.destinationFile}\n\n` +
                    `Please verify that the target directory has the correct permissions and try again.\n`;
                throw new WriteError(message);
            }
        }
        const publishEndTime = Date.now();
        let assets = [];
        let timeToCopyAssets = 0;
        if (opts.copyAssets) {
            const result = await this.copyAssets();
            assets = result.assets;
            timeToCopyAssets = result.timeToCopy;
        }
        const publishResult = {
            pages: publishedPages,
            assets,
            timeToPublish: publishEndTime - publishStartTime,
            timeToCopyAssets
        };
        for await (const plugin of this.config.plugins) {
            if (plugin.afterAll) {
                try {
                    await plugin.afterAll(publishResult, this.config);
                }
                catch (err) {
                    const message = `A plugin called "${plugin.name} threw an error during the "afterAll" phase.\n\n` +
                        `Here's some more info about the error.\n\n` +
                        `${err.message}\n`;
                    throw new AfterAllPluginError(plugin.name, message);
                }
            }
        }
        return publishResult;
    }
    async copyAssets() {
        if (this.config.assets) {
            const startTime = Date.now();
            const from = `${path.join(this.config.source, this.config.assets)}/**/*`;
            const to = path.join(this.config.destination, this.config.assets);
            try {
                await fs.mkdir(to, { recursive: true });
                const assets = await cpy(from, to);
                const timeToCopy = Date.now() - startTime;
                return {
                    assets,
                    timeToCopy
                };
            }
            catch (err) {
                const message = `Doxicity wasn't able to copy assets from "${from}" to "${to}".\n\n` +
                    `Please verify that the target directories have the correct permissions and try again.\n`;
                throw new WriteError(message);
            }
        }
        return {
            assets: [],
            timeToCopy: 0
        };
    }
    async clean() {
        await fs.rm(this.config.destination, { recursive: true, force: true });
    }
}
