import fs from 'fs/promises';
import path from 'path';
import lunr from 'lunr';
import { getRelativeUrl } from '../../utilities/files.js';
const map = [];
function stripWhitespace(string) {
    return string.replace(/\s+/g, ' ');
}
export default function (options) {
    const opts = {
        searchBoxContainer: '#search-box',
        ignore: () => false,
        getTitle: doc => doc.querySelector('title')?.textContent?.trim() ?? '',
        getDescription: doc => doc.querySelector('meta[name="description"]')?.getAttribute('content') ?? '',
        getHeadings: doc => [...doc.querySelectorAll('h2, h3, h4')].map(heading => heading.textContent?.trim() ?? ''),
        getContent: doc => doc.querySelector('body')?.textContent?.trim() || '',
        directoryName: 'search',
        ...options
    };
    let pagesToIndex = [];
    return {
        name: 'doxicity-search-plugin',
        scripts: async (config) => {
            const clientSource = await fs.readFile('./src/plugins/search/search-client.js', 'utf8');
            const lunrSource = await fs.readFile('./node_modules/lunr/lunr.min.js', 'utf8');
            const searchIndexFilename = path.join(config.destination, config.assets, path.join(opts.directoryName, 'search-index.json'));
            const searchIndexUrl = getRelativeUrl(config, searchIndexFilename);
            const scripts = `
        /* prettier-ignore */
        ${lunrSource}

        ${clientSource.replace(/SEARCH_INDEX_URL/g, searchIndexUrl)}
      `;
            return scripts;
        },
        styles: async () => {
            return await fs.readFile('./src/plugins/search/search-client.css', 'utf8');
        },
        transform: (doc, page, config) => {
            const searchContainer = doc.querySelector(opts.searchBoxContainer);
            const searchBox = doc.createElement('button');
            searchBox.classList.add('site-search-box');
            searchBox.type = 'button';
            searchBox.title = 'Press / to search';
            searchBox.setAttribute('aria-label', 'Search');
            searchBox.setAttribute('data-plugin', 'search');
            searchBox.innerHTML = `
        <svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" viewBox="0 0 16 16" aria-hidden="true">
          <path d="M11.742 10.344a6.5 6.5 0 1 0-1.397 1.398h-.001c.03.04.062.078.098.115l3.85 3.85a1 1 0 0 0 1.415-1.414l-3.85-3.85a1.007 1.007 0 0 0-.115-.1zM12 6.5a5.5 5.5 0 1 1-11 0 5.5 5.5 0 0 1 11 0z"></path>
        </svg>
        <span>Search</span>
      `;
            searchContainer?.append(searchBox);
            const title = stripWhitespace(opts.getTitle(doc));
            const description = stripWhitespace(opts.getDescription(doc));
            const headings = opts.getHeadings(doc).map(stripWhitespace);
            const content = stripWhitespace(opts.getContent(doc));
            const url = getRelativeUrl(config, page.destinationFile);
            pagesToIndex.push({ page, title, description, headings, content, url });
            return doc;
        },
        afterAll: (_pages, config) => {
            const searchIndexFilename = path.join(config.destination, config.assets, path.join(opts.directoryName, 'search-index.json'));
            const searchIndex = lunr(async function () {
                let index = 0;
                this.ref('id');
                this.field('t', { boost: 10 });
                this.field('h', { boost: 5 });
                this.field('c');
                for (const page of pagesToIndex) {
                    if (opts.ignore(page.page))
                        continue;
                    this.add({ id: index, t: page.title, h: page.headings, c: page.content });
                    map[index] = { title: page.title, description: page.description, url: page.url };
                    index++;
                }
                await fs.mkdir(path.dirname(searchIndexFilename), { recursive: true });
                await fs.writeFile(searchIndexFilename, JSON.stringify({ searchIndex, map }), 'utf8');
                pagesToIndex = [];
            });
        }
    };
}
